// ChangeTest.cpp : This file contains the 'main' function. Program execution begins and ends there.
//

#include <windows.h>
#include <conio.h>
#include <stdio.h>
#include <string>

class NotifyWatcher
{
	OVERLAPPED m_ov{};
	HANDLE m_hDir{};
	HANDLE m_hWaitEvent{};
	BYTE *m_pBuffer{};
	bool m_fValid{};
	bool m_fRecurse{};
	std::wstring m_strDir;

	const DWORD dwNotifyFilter{
		(FILE_NOTIFY_CHANGE_FILE_NAME
		|FILE_NOTIFY_CHANGE_DIR_NAME
		|FILE_NOTIFY_CHANGE_ATTRIBUTES
		|FILE_NOTIFY_CHANGE_LAST_WRITE
		|FILE_NOTIFY_CHANGE_CREATION
		|FILE_NOTIFY_CHANGE_SIZE)};

	const DWORD dwBuffSize{32768};

	static VOID WINAPI notifyRoutine(DWORD dwErrorCode, DWORD dwNumberOfBytesTransfered, LPOVERLAPPED lpOverlapped)
	{
		NotifyWatcher *pThis = reinterpret_cast< NotifyWatcher * >( lpOverlapped->hEvent );

		const wchar_t *szIndent = L"    ";

		wprintf(L"Notified with %ld bytes.\n", dwNumberOfBytesTransfered);

		if (dwNumberOfBytesTransfered == 0)
		{
			wprintf(szIndent);
			wprintf(L"Notification OVERFLOW. (Can be normal under extreme load.)\n");
		}
		else if (dwNumberOfBytesTransfered < sizeof(FILE_NOTIFY_INFORMATION))
		{
			wprintf(szIndent);
			wprintf(L"Notification size INVALID.\n");
		}
		else
		{
			FILE_NOTIFY_INFORMATION *pInfo = reinterpret_cast< FILE_NOTIFY_INFORMATION * >(pThis->m_pBuffer);

			while(true)
			{
				const wchar_t *szAction{};
				switch(pInfo->Action)
				{
				default: break;
				case FILE_ACTION_ADDED:            szAction = L"FILE_ACTION_ADDED           "; break;
				case FILE_ACTION_REMOVED:          szAction = L"FILE_ACTION_REMOVED         "; break;
				case FILE_ACTION_MODIFIED:         szAction = L"FILE_ACTION_MODIFIED        "; break;
				case FILE_ACTION_RENAMED_OLD_NAME: szAction = L"FILE_ACTION_RENAMED_OLD_NAME"; break;
				case FILE_ACTION_RENAMED_NEW_NAME: szAction = L"FILE_ACTION_RENAMED_NEW_NAME"; break;
				}

				wprintf(szIndent);
				if (szAction)
				{
					wprintf(L"%s", szAction);
				}
				else
				{
					wprintf(L"Unknown action %-13d", pInfo->Action);
				}
				wprintf(szIndent);
				if (pInfo->FileNameLength == 0 || ((pInfo->FileNameLength)/sizeof(wchar_t)) > INT_MAX)
				{
					wprintf(L"Invalid filename length!");
				}
				else
				{
					int nameLen = ((pInfo->FileNameLength)/sizeof(wchar_t));
					wprintf(L"\"%s%.*s\"", pThis->m_strDir.c_str(), nameLen, pInfo->FileName);
				}
				wprintf(L"\n");

				if (pInfo->NextEntryOffset == 0)
				{
					break;
				}

				pInfo = reinterpret_cast< FILE_NOTIFY_INFORMATION * >(reinterpret_cast< BYTE * >(pInfo) + pInfo->NextEntryOffset);
			}
		}

		// Ask for more.
		pThis->Schedule();
	}

	bool Schedule()
	{
		if (!m_fValid)
		{
			return false;
		}

		DWORD dwUnused{};

		if (!::ReadDirectoryChangesW(m_hDir, m_pBuffer, dwBuffSize, m_fRecurse?TRUE:FALSE, dwNotifyFilter, &dwUnused, &m_ov, notifyRoutine))
		{
			DWORD dwErr = ::GetLastError();
			wprintf(L"ReadDirectoryChangesW failed on \"%s\": %ld\n", m_strDir.c_str(), dwErr);
			return false;
		}
		return true;
	}

public:
	NotifyWatcher(const wchar_t *szDir, bool fRecurse)
		: m_strDir(szDir)
		, m_fRecurse(fRecurse)
	{
		if (!m_strDir.empty() && m_strDir.back() != L'\\')
		{
			m_strDir += L'\\';
		}

		ZeroMemory(&m_ov,sizeof(m_ov));
		// ReadDirectoryChangesW: When using a completion routine, "the hEvent member of the OVERLAPPED structure is not
		// used by the system, so you can use it yourself."
		m_ov.hEvent = reinterpret_cast< HANDLE >( this );

		m_hDir = ::CreateFile(
			m_strDir.c_str(),
			FILE_LIST_DIRECTORY,
			FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,
			0,
			OPEN_EXISTING,
			FILE_FLAG_BACKUP_SEMANTICS|FILE_FLAG_OVERLAPPED,
			0);

		m_hWaitEvent = ::CreateEvent(nullptr, TRUE, FALSE, nullptr);

		if (m_hDir == INVALID_HANDLE_VALUE)
		{
			DWORD dwErr = ::GetLastError();
			wprintf(L"Failed to open \"%s\": %ld\n", m_strDir.c_str(), dwErr);
		}
		else if (!m_hWaitEvent)
		{
			DWORD dwErr = ::GetLastError();
			wprintf(L"Failed to create event: %ld\n", dwErr);
		}
		else
		{
			m_pBuffer = new BYTE[dwBuffSize];
			m_fValid = true;

			if (Schedule())
			{
				wprintf(L"Watching \"%s\" (%srecursive)\nPush any key to exit.\n", m_strDir.c_str(), m_fRecurse?L"":L"non-");
			}
		}
	}

	~NotifyWatcher()
	{
		if (m_hDir != INVALID_HANDLE_VALUE)
		{
			::CloseHandle(m_hDir);
			m_hDir = INVALID_HANDLE_VALUE;
		}

		if (m_hWaitEvent)
		{
			::CloseHandle(m_hWaitEvent);
			m_hWaitEvent = nullptr;
		}

		delete[] m_pBuffer;
	}

	bool Wait(DWORD dwTimeout)
	{
		if (!m_fValid)
		{
			wprintf(L"Failed.\n");
			return false;
		}

		// Nothing ever sets m_hWaitEvent, but we can't specify zero events.
		// We wait so the completion event has a chance to be called.
		DWORD dwWaitRes = ::WaitForMultipleObjectsEx(1, &m_hWaitEvent, FALSE, dwTimeout, TRUE);

		if (dwWaitRes != WAIT_TIMEOUT && dwWaitRes != WAIT_IO_COMPLETION)
		{
			DWORD dwErr = ::GetLastError();
			wprintf(L"Wait failed: 0x%lx %ld.\n", dwWaitRes, dwErr);
			return false;
		}

		return true;
	}
};

bool IsString(const wchar_t *szStringToTest, const wchar_t *szString1, const wchar_t *szString2 = nullptr, const wchar_t *szString3 = nullptr)
{
	if (!szStringToTest || !szString1)
	{
		return false;
	}

	if (szString1 && _wcsicmp(szStringToTest, szString1) == 0)
	{
		return true;
	}

	if (szString2 && _wcsicmp(szStringToTest, szString2) == 0)
	{
		return true;
	}

	if (szString3 && _wcsicmp(szStringToTest, szString3) == 0)
	{
		return true;
	}

	return false;
}

bool IsHelpString(const wchar_t *sz)
{
	while(*sz == L'-' || *sz == L'/')
	{
		++sz;
	}

	return IsString(sz, L"h", L"help", L"?");
}

int wmain(int argc, wchar_t **argv)
{
	bool fShowHelp{};
	bool fRecurse{true};
	std::wstring strPath;

	for(int iArg = 1; iArg < argc; ++iArg)
	{
		if (_wcsicmp(argv[iArg], L"NoRecurse") == 0)
		{
			fRecurse = false;
		}
		else if (IsHelpString(argv[iArg]))
		{
			fShowHelp = true;
		}
		else if (!strPath.empty())
		{
			fShowHelp = true;
			wprintf(L"\nToo many arguments. (Paths containing spaces must go in quotes.)\n");
		}
		else
		{
			strPath = argv[iArg];
		}
	}

	if (fShowHelp)
	{
		wprintf(L"\n");
		wprintf(L"Run without any arguments to monitor C:\\ recursively.\n");
		wprintf(L"Add the \"NoRecurse\" argument to monitor only the target and not sub-folders.\n");
		wprintf(L"Specify a folder path to monitor that instead of C:\\.\n");
		wprintf(L"Examples:\n");
		wprintf(L"    ChangeText.exe D:\\Test\n");
		wprintf(L"    ChangeText.exe C:\\ NoRecurse\n");
		wprintf(L"    ChangeText.exe \"C:\\Program Files\"\n");
		return -1;
	}

	if (strPath.empty())
	{
		strPath = L"C:\\";
	}

	NotifyWatcher watcher{ strPath.c_str(), fRecurse };

	// Poll for keypress every 200ms while in an alertable state.
	while(!_kbhit())
	{
		if (!watcher.Wait(200))
		{
			return -1;
		}
	}

	return 0;
}
